from pydantic import BaseModel, Field, field_validator

from typing import Any, Optional
from datetime import datetime
import logging

logger = logging.getLogger("OPGG.py")


class ChampionStats(BaseModel):
    id: Optional[int] = None
    play: Optional[int] = None
    win: Optional[int] = None
    lose: Optional[int] = None
    game_length_second: Optional[int] = None
    kill: Optional[int] = None
    death: Optional[int] = None
    assist: Optional[int] = None
    gold_earned: Optional[int] = None
    minion_kill: Optional[int] = None
    neutral_minion_kill: Optional[int] = None
    damage_taken: Optional[int] = None
    damage_dealt_to_champions: Optional[int] = None
    double_kill: Optional[int] = 0
    triple_kill: Optional[int] = 0
    quadra_kill: Optional[int] = 0
    penta_kill: Optional[int] = 0
    vision_wards_bought_in_game: Optional[int] = 0
    op_score: Optional[int] = None
    snowball_throws: Optional[int] = None
    snowball_hits: Optional[int] = None

    @field_validator("*", mode="after")
    def log_none_values(cls, v, info):
        if v is None:
            logger.warning(f"Field '{info.field_name}' is None in ChampionStats model")
        return v

    @property
    def winrate(self) -> int:
        """`[Computed Property]` Returns the winrate percentage of the champion."""
        return round(self.win / self.play * 100) if self.play and self.win else 0


class MostChampions(BaseModel):
    game_type: Optional[str] = None
    season_id: Optional[int] = None
    year: Optional[int] = None
    play: Optional[int] = None
    win: Optional[int] = None
    lose: Optional[int] = None
    champion_stats: Optional[list[ChampionStats]] = Field(default_factory=list)

    @field_validator("*", mode="after")
    def log_none_values(cls, v, info):
        if v is None:
            logger.warning(f"Field '{info.field_name}' is None in MostChampions model")
        return v


class Price(BaseModel):
    currency: Optional[str] = None
    cost: Optional[int] = None

    @field_validator("*", mode="after")
    def log_none_values(cls, v, info):
        if v is None:
            logger.warning(f"Field '{info.field_name}' is None in Price model")
        return v


class Skin(BaseModel):
    id: Optional[int] = None
    champion_id: Optional[int] = None
    name: Optional[str] = None
    has_chromas: Optional[bool] = None
    splash_image: Optional[str] = None
    loading_image: Optional[str] = None
    tiles_image: Optional[str] = None
    centered_image: Optional[str] = None
    skin_video_url: Optional[str] = None
    prices: Optional[list[Price]] = None
    sales: Optional[Any] = None
    release_date: Optional[datetime] = None

    @field_validator("*", mode="after")
    def log_none_values(cls, v, info):
        if v is None:
            logger.warning(f"Field '{info.field_name}' is None in Skin model")
        return v


class Info(BaseModel):
    attack: Optional[int] = None
    defense: Optional[int] = None
    magic: Optional[int] = None
    difficulty: Optional[int] = None

    @field_validator("*", mode="after")
    def log_none_values(cls, v, info):
        if v is None:
            logger.warning(f"Field '{info.field_name}' is None in Info model")
        return v


class Stats(BaseModel):
    hp: Optional[float] = None
    hpperlevel: Optional[float] = None
    mp: Optional[float] = None
    mpperlevel: Optional[float] = None
    movespeed: Optional[float] = None
    armor: Optional[float] = None
    armorperlevel: Optional[float] = None
    spellblock: Optional[float] = None
    spellblockperlevel: Optional[float] = None
    attackrange: Optional[float] = None
    hpregen: Optional[float] = None
    hpregenperlevel: Optional[float] = None
    mpregen: Optional[float] = None
    mpregenperlevel: Optional[float] = None
    crit: Optional[float] = None
    critperlevel: Optional[float] = None
    attackdamage: Optional[float] = None
    attackdamageperlevel: Optional[float] = None
    attackspeed: Optional[float] = None
    attackspeedperlevel: Optional[float] = None

    @field_validator("*", mode="after")
    def log_none_values(cls, v, info):
        if v is None:
            logger.warning(f"Field '{info.field_name}' is None in Stats model")
        return v


class Passive(BaseModel):
    name: Optional[str] = None
    description: Optional[str] = None
    image_url: Optional[str] = None
    video_url: Optional[str] = None

    @field_validator("*", mode="after")
    def log_none_values(cls, v, info):
        if v is None:
            logger.warning(f"Field '{info.field_name}' is None in Passive model")
        return v


class Spell(BaseModel):
    key: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None
    max_rank: Optional[int] = None
    range_burn: Optional[list[int]] = Field(default_factory=list)
    cooldown_burn: Optional[list[float]] = Field(default_factory=list)
    cooldown_burn_float: Optional[list[float]] = Field(default_factory=list)
    cost_burn: Optional[list[int]] = Field(default_factory=list)
    tooltip: Optional[str] = None
    image_url: Optional[str] = None
    video_url: Optional[str] = None

    @field_validator("*", mode="after")
    def log_none_values(cls, v, info):
        if v is None:
            logger.warning(f"Field '{info.field_name}' is None in Spell model")
        return v


class Evolve(BaseModel):
    key: Optional[str] = None
    name: Optional[str] = None
    image_url: Optional[str] = None

    @field_validator("*", mode="after")
    def log_none_values(cls, v, info):
        if v is None:
            logger.warning(f"Field '{info.field_name}' is None in Evolve model")
        return v


class Champion(BaseModel):
    id: Optional[int] = None
    key: Optional[str] = None
    name: Optional[str] = None
    image_url: Optional[str] = None
    evolve: Optional[list[Evolve]] = None
    blurb: Optional[str] = None
    title: Optional[str] = None
    tags: Optional[list[str]] = Field(default_factory=list)
    lore: Optional[str] = None
    partype: Optional[str] = None
    info: Optional[Info] = None
    stats: Optional[Stats] = None
    enemy_tips: Optional[list[str]] = Field(default_factory=list)
    ally_tips: Optional[list[str]] = Field(default_factory=list)
    skins: Optional[list[Skin]] = Field(default_factory=list)
    passive: Optional[Passive] = None
    spells: Optional[list[Spell]] = Field(default_factory=list)

    @field_validator("*", mode="after")
    def log_none_values(cls, v, info):
        if v is None:
            logger.warning(f"Field '{info.field_name}' is None in Champion model")
        return v



from pydantic import BaseModel, field_validator
import logging

logger = logging.getLogger("OPGG.py")


class OPScore(BaseModel):
    second: int
    score: float

    @field_validator("*", mode="after")
    def log_none_values(cls, v, info):
        if v is None:
            logger.warning(f"Field '{info.field_name}' is None in OPScore model")
        return v


class OPScoreAnalysis(BaseModel):
    left: str
    right: str
    last: str

    @field_validator("*", mode="after")
    def log_none_values(cls, v, info):
        if v is None:
            logger.warning(
                f"Field '{info.field_name}' is None in OPScoreAnalysis model"
            )
        return v



import os
import logging
import asyncio
import pprint

from datetime import datetime
from typing import Literal
from fake_useragent import UserAgent

from opgg.v2.champion import Champion
from opgg.v2.game import Game
from opgg.v2.summoner import Summoner
from opgg.v2.response import LiveGameResponse, UpdateResponse
from opgg.v2.utils import Utils
from opgg.v2.cacher import Cacher
from opgg.v2.params import By, LangCode, Region, GenericReqParams
from opgg.v2.search_result import SearchResult


class OPGG:
    """
    ### OPGG.py

    Copyright (c) 2023-2025, ShoobyDoo

    License: BSD-3-Clause, See LICENSE for more details.
    """

    __author__ = "ShoobyDoo"
    __license__ = "BSD-3-Clause"

    def __init__(self) -> None:
        self.BYPASS_API_URL = "https://lol-web-api.op.gg/api/v1.0/internal/bypass"
        self.SEARCH_API_URL = f"{self.BYPASS_API_URL}/summoners/v2/{{region}}/autocomplete?gameName={{summoner_name}}&tagline={{tagline}}"
        self.SUMMARY_API_URL = (
            f"{self.BYPASS_API_URL}/summoners/{{region}}/{{summoner_id}}/summary"
        )
        self.GAMES_API_URL = f"{self.BYPASS_API_URL}/games/{{region}}/summoners/{{summoner_id}}?&limit={{limit}}&game_type={{game_type}}&hl={{hl}}"
        self.RENEW_API_URL = (
            f"{self.BYPASS_API_URL}/summoners/{{region}}/{{summoner_id}}/renewal"
        )
        self.CHAMPIONS_API_URL = f"{self.BYPASS_API_URL}/meta/champions?hl={{hl}}"
        self.CHAMPION_BY_ID_API_URL = (
            f"{self.BYPASS_API_URL}/meta/champions/{{champion_id}}?hl={{hl}}"
        )
        self.LIVE_GAME_URL = (
            f"{self.BYPASS_API_URL}/spectates/{{region}}/{{summoner_id}}?hl={{hl}}"
        )

        # FIGURE OUT HOW TO FIND THESE PAYLOADS:
        # - seasonById
        # - spells

        self._ua = UserAgent()
        self._headers = {"User-Agent": self._ua.random}

        # ===== SETUP START =====
        logging.root.name = "OPGG.py"

        if not os.path.exists("./logs"):
            logging.info("Creating logs directory...")
            os.mkdir("./logs")
        else:
            # remove empty log files
            for file in os.listdir("./logs"):
                if (
                    os.stat(f"./logs/{file}").st_size == 0
                    and file != f'opgg_{datetime.now().strftime("%Y-%m-%d")}.log'
                ):
                    logging.info(f"Removing empty log file: {file}")
                    os.remove(f"./logs/{file}")

        logging.basicConfig(
            filename=f'./logs/opgg_{datetime.now().strftime("%Y-%m-%d")}.log',
            filemode="a+",
            format="[%(asctime)s][%(name)s->%(module)-15s:%(lineno)10d][%(levelname)-7s] : %(message)s",
            datefmt="%d-%b-%y %H:%M:%S",
            level=logging.INFO,
        )
        # ===== SETUP END =====

        # allow the user to interact with the logger
        self._logger = logging.getLogger("OPGG.py")

        self.logger.info(
            f"OPGG.__init__(BYPASS_API_URL={self.BYPASS_API_URL}, "
            f"headers={self._headers})"
        )

        # at object creation, setup and query the cache
        self._cacher = Cacher()
        self._cacher.setup()

    @property
    def logger(self) -> logging.Logger:
        """
        A `Logger` object representing the logger instance.

        The logging level is set to `INFO` by default.
        """
        return self._logger

    @property
    def headers(self) -> dict:
        """
        A `dict` representing the headers to send with the request.
        """
        return self._headers

    @headers.setter
    def headers(self, value: dict) -> None:
        self._headers = value

    # /AUTOCOMPLETE
    def search(
        self,
        query: str,
        region: Region = Region.ANY,
        returns: Literal["simple", "full"] = "full",
    ) -> list[SearchResult]:
        """
        Search for League of Legends summoners by name across one or all regions.

        #### Args:
            - `query` (`str`): Summoner name to search for. Can include tagline (e.g., "name#tag")
            - `region` (`Region`, optional): Specific region to search in. Defaults to `Region.ANY` which searches all regions concurrently.

        #### Returns:
            - `list[SearchResult]`: List of `SearchResult` objects containing the found summoners.
              Each `SearchResult` object contains formatted summoner data and the region it was found in.

        #### Example:
        ```python
        >>> opgg = OPGG()
        >>>
        >>> results = opgg.search("faker", Region.KR) # Search specific region (recommended)
        >>> all_results = opgg.search("faker") # Search all regions (resource-intensive)
        >>>
        >>> [print(result) for result in results]
        >>> [print(result) for result in all_results]
        ```

        #### Note:
        Searching with `Region.ANY` is significantly more resource-intensive as it
        queries all regional endpoints concurrently. Additionally, when searching all
        regions, there is a hard limit of 10 search results per region.
        """

        self.logger.info(f"Starting summoner search for '{query}' in region '{region}'")

        search_params: GenericReqParams = {
            "base_api_url": self.SEARCH_API_URL,
            "headers": self.headers,
        }

        if region == Region.ANY:
            self.logger.info("Searching all regions...")
            results = asyncio.run(Utils._search_all_regions(query, search_params))
        else:
            self.logger.info(f"Searching specific region: {region}")
            results = asyncio.run(
                Utils._single_region_search(
                    query,
                    region,
                    search_params,
                )
            )

        # Inner object instantiation BEFORE SearchResult wrapper
        search_results = [
            SearchResult(
                **{
                    "summoner": Summoner(**result["summoner"]),
                    "region": result["region"],
                }
            )
            for result in results
        ]

        if returns == "profile":
            # set each summoner property in the search_results array
            # to the corresponding Summoner object
            self.logger.info(
                f"Fetching profiles for {len(search_results)} summoners..."
            )

            # This call is done passing the array because if there are multiple, we want the benefit of concurrency
            # provided by the asychonous nature of the call
            summoners = self.get_summoner(search_results)

            # We then build the search_results array with the Summoner objects
            for sr, s in zip(search_results, summoners):
                sr.summoner = s

        return search_results

    def get_summoner(
        self,
        search_result: SearchResult | list[SearchResult] = None,
        summoner_id: str | list[str] = None,
        region: Region | list[Region] = None,
    ) -> Summoner | list[Summoner]:
        if search_result is None and summoner_id is not None and region is not None:
            if isinstance(summoner_id, str) and isinstance(region, Region):
                search_result = SearchResult(
                    **{
                        "summoner": Summoner(**{"summoner_id": summoner_id}),
                        "region": region,
                    }
                )
            elif isinstance(summoner_id, list) and isinstance(region, list):
                search_result = [
                    SearchResult(
                        **{"summoner": Summoner(**{"summoner_id": sid}), "region": reg}
                    )
                    for sid, reg in zip(summoner_id, region)
                ]
            else:
                raise ValueError("Mismatched types for summoner_id and region")

        if isinstance(search_result, SearchResult):
            self.logger.info(f"Fetching profile for summoner result: {search_result}")
            profile_data = asyncio.run(
                Utils._fetch_profile(
                    search_result.summoner.summoner_id,
                    params={
                        "base_api_url": self.SUMMARY_API_URL.format_map(
                            {
                                "region": search_result.region,
                                "summoner_id": search_result.summoner.summoner_id,
                            }
                        ),
                        "headers": self.headers,
                    },
                )
            )
            return Summoner(**profile_data["summoner"])

        elif isinstance(search_result, list):
            self.logger.info(
                f"Fetching profiles for {len(search_result)} summoner results"
            )
            self.logger.debug(
                f"Raw summoner results: \n{pprint.pformat(search_result)}"
            )
            profile_data = asyncio.run(
                Utils._fetch_profile_multiple(
                    {
                        "base_api_url": self.SUMMARY_API_URL,
                        "headers": self.headers,
                    },
                    search_result,
                )
            )
            return [Summoner(**profile["summoner"]) for profile in profile_data]

        else:
            raise ValueError("Invalid type for search_result")

    def get_recent_games(
        self,
        search_result: SearchResult | list[SearchResult] = None,
        summoner_id: str | list[str] = None,
        region: Region | list[Region] = None,
        results: int = 15,
        game_type: Literal["total", "ranked", "normal"] = "total",
        lang_code=LangCode.ENGLISH,
    ) -> list[Game] | list[list[Game]]:
        """
        Retrieve recent games for the given summoner(s).

        #### Args:
            - `search_result` (`SearchResult` or `list[SearchResult]`, optional): The search result(s) representing the summoner(s) whose recent games are to be retrieved. If not provided, you must supply `summoner_id` and `region`.
            - `summoner_id` (`str` or `list[str]`, optional): Summoner ID(s) for which to fetch recent games (used when `search_result` is not provided).
            - `region` (`Region` or `list[Region]`, optional): The region(s) corresponding to the summoner ID(s).
            - `results` (`int`, optional): The number of recent games to retrieve. Defaults to 15.
            - `game_type` (`str`, optional): The type of games to retrieve (e.g., "total", "ranked", "normal"). Defaults to "total".
            - `lang_code` (`LangCode`, optional): Language code for localization. Defaults to `LangCode.ENGLISH`.

        #### Returns:
        - If a single `SearchResult` is provided:
            `list[Game]`: A list of `Game` objects representing the recent games for that summoner.
        - If a list of `SearchResult` objects is provided:
            `list[list[Game]]`: A list of lists, where each inner list contains `Game` objects corresponding to each individual `SearchResult`.

        #### Example:
        ```python
        >>> opgg = OPGG()
        >>> # For a single summoner:
        >>> games = opgg.get_recent_games(search_result=some_search_result)
        >>> # For multiple summoners:
        >>> grouped_games = opgg.get_recent_games(search_result=[result1, result2])
        ```
        """
        
        if search_result is None and summoner_id is not None and region is not None:
            if isinstance(summoner_id, str) and isinstance(region, Region):
                search_result = SearchResult(
                    {
                        "summoner": Summoner(**{"summoner_id": summoner_id}),
                        "region": region,
                    }
                )
            elif isinstance(summoner_id, list) and isinstance(region, list):
                search_result = [
                    SearchResult(
                        {"summoner": Summoner(**{"summoner_id": sid}), "region": reg}
                    )
                    for sid, reg in zip(summoner_id, region)
                ]
            else:
                raise ValueError("Mismatched types for summoner_id and region")

        if isinstance(search_result, SearchResult):
            self.logger.info(
                f'Fetching {results} recent games of type "{game_type}" for summoner result: {search_result}'
            )
            game_params: GenericReqParams = {
                "base_api_url": self.GAMES_API_URL.format_map(
                    {
                        "region": search_result.region,
                        "summoner_id": search_result.summoner.summoner_id,
                        "limit": results,
                        "game_type": game_type,
                        "hl": lang_code,
                    }
                ),
                "headers": self.headers,
            }
            game_data = asyncio.run(
                Utils._fetch_recent_games(
                    game_params,
                )
            )

            return [Game(**game) for game in game_data]

        elif isinstance(search_result, list):
            self.logger.info(
                f'Fetching {results} recent games of type "{game_type}" for {len(search_result)} summoner results'
            )
            game_params_list = [
                {
                    "base_api_url": self.GAMES_API_URL.format_map(
                        {
                            "region": sr.region,
                            "summoner_id": sr.summoner.summoner_id,
                            "limit": results,
                            "game_type": game_type,
                            "hl": lang_code,
                        }
                    ),
                    "headers": self.headers,
                }
                for sr in search_result
            ]
            game_data_list = asyncio.run(
                Utils._fetch_recent_games_multiple(
                    game_params_list,
                )
            )
            return [
                [Game(**game) for game in game_data] for game_data in game_data_list
            ]

        else:
            raise ValueError("Invalid type for search_result")

    def update(self, search_result: SearchResult) -> UpdateResponse:
        """
        Send an update request to fetch the latest details for a given summoner (id).

        #### Note: `This can take several seconds to complete. Assuming there is an update to the profile, this function will not return until the update is complete.`

        ### Parameters
            search_result : `SearchResult`
                Pass a SearchResult object to be updated

        ### Returns
            `UpdateResponse` : Returns an object with the update response payload.

            Example response:
            ```
            {
                'status': 202,
                'data': {
                    'message': 'Already renewed.',
                    'last_updated_at': '2024-07-13T10:06:11+09:00',
                    'renewable_at': '2024-07-13T10:08:12+09:00'
                }
            }
            ```
        """

        update_params: GenericReqParams = {
            "base_api_url": self.RENEW_API_URL,
            "headers": self.headers,
        }

        return asyncio.run(Utils._update(search_result, update_params))

    def get_all_champions(
        self, lang_code: LangCode = LangCode.ENGLISH
    ) -> list[Champion]:
        get_champs_params: GenericReqParams = {
            "base_api_url": self.CHAMPIONS_API_URL.format_map({"hl": lang_code}),
            "headers": self.headers,
        }

        champions_list = asyncio.run(Utils._fetch_all_champions(get_champs_params))
        champions_list_objs = [Champion(**champ) for champ in champions_list]

        self.logger.info(f"Got {len(champions_list_objs)} champions")

        if (len(champions_list_objs)) != self._cacher.get_cached_champs_count():
            self.logger.info(
                "Champ response has more records than are saved. Rebuilding cache..."
            )
            self._cacher.cache_champs(champions_list_objs)

        return champions_list_objs

    def get_champion_by(
        self, by: By, value: str | int, lang_code: LangCode = LangCode.ENGLISH
    ) -> Champion | list[Champion]:
        if by == By.ID:
            get_champ_by_id_params: GenericReqParams = {
                "base_api_url": self.CHAMPION_BY_ID_API_URL,
                "headers": self.headers,
            }

            champ = asyncio.run(
                Utils._fetch_champion_by_id(value, get_champ_by_id_params, lang_code)
            )

        elif by == By.NAME:
            self.logger.info(f"Fetching {value} from cache...")

            champ = self._cacher.get_champ_id_by_name(value)

            if champ == None:
                self.logger.info(f"Champion {value} not found in cache. Fetching...")

                all_champs = self.get_all_champions()

                # Do a loose search by name
                champs = [
                    champ for champ in all_champs if value.lower() in champ.name.lower()
                ]

                if len(champs) == 0:
                    raise ValueError(f"Champion {value} not found")

                if len(champs) > 1:
                    return [Champion(**champ) for champ in champs]

                return Champion(**champs[0])

            else:
                get_champ_by_id_params: GenericReqParams = {
                    "base_api_url": self.CHAMPION_BY_ID_API_URL,
                    "headers": self.headers,
                }

                champ = asyncio.run(
                    Utils._fetch_champion_by_id(
                        champ, get_champ_by_id_params, lang_code
                    )
                )

        else:
            raise ValueError(f"Invalid value for by: {by}")

        return Champion(**champ)

    def get_live_game(
        self, search_result: SearchResult, lang_code: LangCode = LangCode.ENGLISH
    ) -> LiveGameResponse:
        """
        Get the live game data for a given summoner.

        ### Parameters
            search_result : `SearchResult`
                Pass a SearchResult object to get the live game data

        ### Returns
            `LiveGameResponse` : Returns an object with the live game data payload if available.
        """

        self.logger.info(f"Fetching live game data for {search_result}")

        live_game_params: GenericReqParams = {
            "base_api_url": self.LIVE_GAME_URL.format_map(
                {
                    "region": search_result.region,
                    "summoner_id": search_result.summoner.summoner_id,
                    "hl": lang_code,
                }
            ),
            "headers": self.headers,
        }

        live_game_data = asyncio.run(Utils._fetch_live_game(live_game_params))

        return LiveGameResponse(**live_game_data)



# Quick and simple scraper to pull some data from OPGG using multisearch

# Author  : ShoobyDoo
# Date    : 2024-07-10
# License : BSD-3-Clause

from datetime import datetime
import json
import requests
from bs4 import BeautifulSoup
from fake_useragent import UserAgent

from opgg.v1.cacher import Cacher
from opgg.v1.champion import Champion, Passive, Price, Skin, Spell
from opgg.v1.params import By, Region
from opgg.v1.season import SeasonInfo


class Utils:
    """
    ### utils.py
    A collection of static utility helper methods that perform various opgg and league specific tasks such as fetching champions, seasons, etc.\n
    
    Copyright (c) 2023-2024, ShoobyDoo
    License: BSD-3-Clause, See LICENSE for more details.
    """
    
    _base_api_url = "https://lol-web-api.op.gg/api/v1.0/internal/bypass"
    _api_url = f"{_base_api_url}/summoners/{{region}}/{{summoner_id}}/renewal"
    
    ua = UserAgent()
    headers = { 
        "User-Agent": ua.random
    }
    
    @staticmethod
    def update(summoner_id: str, region: Region = Region.NA) -> dict:
        """
        Send an update request to fetch the latest details for a given summoner (id).
        
        ### Parameters
            summoner_id : `str`
                Pass a summoner id as a string to be updated
            
            region : `Region, optional`
                Pass the region you want to perform the update in. Default is "NA".

        ### Returns
            `dict` : Returns a dictionary with the status response.
                Example response:
            ```
            {
                'status': 202, 
                'data': {
                    'message': 'Already renewed.', 
                    'last_updated_at': '2024-07-13T10:06:11+09:00', 
                    'renewable_at': '2024-07-13T10:08:12+09:00'
                }
            }
            ```
        """
        
        res = requests.post(
            Utils._api_url.format(region=region, summoner_id=summoner_id), 
            headers=Utils.headers
        )
        
        if res.status_code in [201, 202]:
            return res.json()
        else:
            res.raise_for_status()
    
    
    @staticmethod
    def get_page_props(summoner_names: str | list[str] = "ColbyFaulkn1", region = Region.NA) -> dict:
        """
        Get the page props from OPGG. (Contains data such as summoner info, champions, seasons, etc.)
        
        ### Parameters
            summoner_names : `str | list[str], optional`
                Pass a single or comma separated `str` or a list of summoner names.\n
                Note: Default is "abc", as this can be any valid summoner if you just want page props. (All champs, seasons, etc.)
            
            region : `Region, optional`
                Pass the region you want to search in. Default is "NA".

        ### Returns
            `dict` : Returns a dictionary with the page props.
        """
        
        if isinstance(summoner_names, list) and len(summoner_names) > 0: 
            summoner_names = ",".join(summoner_names)
        
        url = f"https://www.op.gg/multisearch/{region}?summoners={summoner_names}"

        res = requests.get(url, headers=Utils.headers, allow_redirects=True)
        soup = BeautifulSoup(res.content, "html.parser")
        
        return json.loads(soup.select_one("#__NEXT_DATA__").text)['props']['pageProps']
    
    
    @staticmethod
    def get_all_seasons(region = Region.NA, page_props = None) -> list[SeasonInfo]:
        """
        Get all seasons from OPGG.

        ### Args:
            region : `Region, optional`
                Pass the region you want to search in. Defaults to "NA".
                
            page_props : `dict, optional`
                Pass the page props if the program has queried them once before.\n
                Note: Defaults to None, but if you pass them it reduces the overhead of another request out to OPGG.

        ### Returns:
            `list[SeasonInfo]` : A list of SeasonInfo objects.
        """
        # Check cache, if found, return it, otherwise continue to below logic.
        cached_seasons = Cacher().get_all_seasons()
        if cached_seasons:
            return cached_seasons
        
        seasons = []
        
        # For seasons specifically, if page_props is not passed, we MUST use it.
        # I have not been able to find a seasons endpoint on the api yet.
        if page_props == None:
            page_props = Utils.get_page_props()
        
        for season in dict(page_props['seasonsById']).values():
            seasons.append(SeasonInfo(
                id = season["id"],
                value = season["value"],
                display_value = season["display_value"],
                split = season["split"],
                is_preseason = season["is_preseason"]
            ))
        
        return seasons


    @staticmethod
    def get_season_by(by: By, value: int | str | list) -> SeasonInfo | list[SeasonInfo]:
        """
        Get a season by a specific metric.
        
        ### Args:
            by : `By`
                Pass a By enum to specify how you want to get the season(s).
            
            value : `int | str | list`
                Pass the value(s) you want to search by. (id, display_value, etc.)
        
        ### Returns:
            `SeasonInfo | list[SeasonInfo]` : A single or list of SeasonInfo objects.
        """
        all_seasons = Utils.get_all_seasons()
        result_set = []
        
        if by == By.ID:
            if isinstance(value, list):
                for season in all_seasons:
                    for id in value:
                        if season.id == id:
                            result_set.append(season)
            else:
                for season in all_seasons:
                    if season.id == int(value):
                        result_set.append(season)
        
        # TODO: perhaps add more ways to get season objs, like by is_preseason, or display_name, etc.           
        
        return result_set if len(result_set) > 1 else result_set[0]
    
    
    @staticmethod
    def get_all_champions(region = Region.NA, page_props = None) -> list[Champion]:
        """
        Get all champion info from OPGG.

        ### Args:
            region : `Region, optional`
                Pass the region you want to search in. Defaults to "NA".
                
            page_props : `dict, optional`
                Pass the page props if the program has queried them once before.\n
                Note: Defaults to None, but if you pass them it reduces the overhead of another request out to OPGG.

        Returns:
            `list[Champion]` : A list of Champion objects.
        """
        # Check cache, if found, return it, otherwise continue to below logic.        
        champions = []
        
        if not page_props:
            cached_champions = Cacher().get_all_champs()
            if cached_champions: 
                return cached_champions
            
            res = requests.get(f"{Utils._base_api_url}/meta/champions?hl=en_US", headers=Utils.headers)
            raw_champs_data = json.loads(res.text)["data"]
            
        else:
            raw_champs_data = dict(page_props['championsById']).values()
        
        for champion in raw_champs_data:
            # reset per iteration
            _spells = []
            _skins = []
            
            for skin in champion["skins"]:
                _prices = []
                
                if skin["prices"]:
                    for price in skin["prices"]:
                        _prices.append(Price(
                            currency = price["currency"] if "RP" in price["currency"] else "BE",
                            cost = price["cost"]
                        ))
                else:
                    _prices = None
                
                _skins.append(Skin(
                    id = skin["id"],
                    champion_id = skin["champion_id"],
                    name = skin["name"],
                    centered_image = skin["centered_image"],
                    skin_video_url = skin["skin_video_url"],
                    prices = _prices,
                    release_date = datetime.fromisoformat(skin["release_date"]) if skin["release_date"] else None,
                    sales = skin["sales"]
                ))
                    
            for spell in champion["spells"]:
                _spells.append(Spell(
                    key = spell["key"],
                    name = spell["name"],
                    description = spell["description"],
                    max_rank = spell["max_rank"],
                    range_burn = spell["range_burn"],
                    cooldown_burn = spell["cooldown_burn"],
                    cooldown_burn_float = spell["cooldown_burn_float"],
                    cost_burn = spell["cost_burn"],
                    tooltip = spell["tooltip"],
                    image_url = spell["image_url"],
                    video_url = spell["video_url"]
                ))       
              
            champions.append(Champion(
                id = champion["id"],
                key = champion["key"],
                name = champion["name"],
                image_url = champion["image_url"],
                evolve = champion["evolve"],
                partype = champion["partype"],
                passive = Passive(
                    name = champion["passive"]["name"],
                    description = champion["passive"]["description"],
                    image_url = champion["passive"]["image_url"],
                    video_url = champion["passive"]["video_url"]
                ),
                spells = _spells,
                skins = _skins
            ))            
        
        return champions
    
    
    @staticmethod
    def get_champion_by(by: By, value: int | str | list, **kwargs) -> Champion | list[Champion]:
        """
        Get a single or list of champions by a specific metric.
        
        ### Args:
            by : `By`
                Pass a By enum to specify how you want to get the champion(s).
                
            value : `int | str | list`
                Pass the value(s) you want to search by. (id, key, name, etc.)
                
            **kwargs : `any`
                Pass any additional keyword arguments to narrow down the search.\n
                Note: Currently only supports "currency" for the cost of a champion.\n
                
                Example: 
                    `get_champion_by(By.COST, 450, currency=By.BLUE_ESSENCE)`
        """
        # Currently kwargs only handles "currency" for the cost of a champion,
        # but I might introduce other metrics of getting champ objs later, idk...
        
        if ("page_props" in kwargs):
            all_champs = Utils.get_all_champions(page_props=kwargs["page_props"])
        else:
            all_champs = Utils.get_all_champions()
        
        result_set = []
        
        if by == By.ID:
            if isinstance(value, list):
                for champ in all_champs:
                    for id in value:
                        if champ.id == id:
                            result_set.append(champ)
            else:
                for champ in all_champs:
                    if champ.id == int(value):
                        result_set.append(champ)
                
        elif by == By.KEY:
            if isinstance(value, list):
                for champ in all_champs:
                    for key in value:
                        if champ.key == key:
                            result_set.append(champ)
            else:
                for champ in all_champs:
                    if champ.key == value:
                        result_set.append(champ)
        
        elif by == By.NAME:
            if isinstance(value, list):
                for champ in all_champs:
                    for name in value:
                        if champ.name == name:
                            result_set.append(champ)
            else:
                for champ in all_champs:
                    if champ.name == value:
                        result_set.append(champ)
        
        elif by == By.COST:
            for champ in all_champs:
                if champ.skins[0].prices:
                    for price in champ.skins[0].prices:
                        if str(kwargs["currency"]).upper() == price.currency and price.cost in value:
                            result_set.append(champ)
                
        
        # if the result set is larger than one, return the whole list, otherwise just return the object itself.
        return result_set if len(result_set) > 1 else result_set[0]


from datetime import datetime
from typing import Optional
from pydantic import BaseModel, HttpUrl, field_validator
import logging

logger = logging.getLogger("OPGG.py")


class QueueInfo(BaseModel):
    id: Optional[int] = None
    queue_translate: Optional[str] = None
    game_type: Optional[str] = None

    @field_validator("*", mode="after")
    def log_none_values(cls, v, info):
        if v is None:
            logger.warning(f"Field '{info.field_name}' is None in QueueInfo model")
        return v


class TierInfo(BaseModel):
    tier: Optional[str] = None
    division: Optional[int] = None
    lp: Optional[int] = None
    level: Optional[int] = None
    tier_image_url: Optional[HttpUrl] = None
    border_image_url: Optional[HttpUrl] = None

    @field_validator("*", mode="after")
    def log_none_values(cls, v, info):
        if v is None:
            logger.warning(f"Field '{info.field_name}' is None in TierInfo model")
        return v


class League(BaseModel):
    game_type: Optional[str] = None
    tier_info: Optional[TierInfo] = None
    win: Optional[int] = None
    lose: Optional[int] = None
    is_hot_streak: Optional[bool] = False
    is_fresh_blood: Optional[bool] = False
    is_veteran: Optional[bool] = False
    is_inactive: Optional[bool] = False
    series: Optional[dict] = None
    updated_at: Optional[datetime] = None

    @field_validator("*", mode="after")
    def log_none_values(cls, v, info):
        if v is None:
            logger.warning(f"Field '{info.field_name}' is None in League model")
        return v

    class Config:
        json_encoders = {datetime: lambda v: v.isoformat()}


class RankInfo(BaseModel):
    """Basic rank information."""

    tier: Optional[str] = None
    division: Optional[int] = None
    lp: Optional[int] = None

    @field_validator("*", mode="after")
    def log_none_values(cls, v, info):
        if v is None:
            logger.warning(f"Field '{info.field_name}' is None in RankInfo model")
        return v


class RankEntry(BaseModel):
    """Entry containing rank information for a specific game type."""

    game_type: str
    rank_info: RankInfo
    created_at: datetime

    class Config:
        json_encoders = {datetime: lambda v: v.isoformat()}


class Season(BaseModel):
    """Represents a League season."""

    season_id: Optional[int] = None
    tier_info: Optional[TierInfo] = None
    created_at: Optional[datetime] = None

    @field_validator("*", mode="after")
    def log_none_values(cls, v, info):
        if v is None:
            logger.warning(f"Field '{info.field_name}' is None in Season model")
        return v

    class Config:
        """Pydantic model configuration."""

        arbitrary_types_allowed = True
        json_encoders = {datetime: lambda v: v.isoformat()}



# Quick and simple scraper to pull some data from OPGG using multisearch

# Author  : ShoobyDoo
# Date    : 2023-07-05
# License : BSD-3-Clause


from datetime import datetime
from typing import Any
from opgg.v1.game import Stats, Team
from opgg.v1.params import By, Queue
from opgg.v1.season import Season
from opgg.v1.league_stats import LeagueStats, QueueInfo, Tier
from opgg.v1.champion import Champion, ChampionStats
from opgg.v1.utils import Utils

# left/right just factor
LJF = 18
RJF = 14


class Participant:
    """
    Represents a participant in the game with detailed information about their performance and loadout.\n
    
    ### Properties:
        `summoner: Summoner` - The summoner associated with this participant\n
        `participant_id: int` - Unique identifier for the participant\n
        `champion_id: int` - Identifier for the champion used by the participant\n
        `team_key: str` - Key representing the participant's team\n
        `position: str` - Position played by the participant (e.g., Top, Mid, Bot)\n
        `role: str` - Role played by the participant (e.g., Carry, Support)\n
        `items: list` - List of items acquired by the participant\n
        `trinket_item: int` - Identifier for the trinket item used by the participant\n
        `rune: dict[str, int]` - Dictionary of runes used by the participant with their respective values\n
        `spells: list` - List of spells used by the participant\n
        `stats: Stats` - Performance statistics of the participant\n
        `tier_info: Tier` - Tier information of the participant\n
    """
    def __init__(self,
                 summoner: 'Summoner',
                 participant_id: int,
                 champion_id: int,
                 team_key: str,
                 position: str,
                 role: str,
                 items: list,
                 trinket_item: int,
                 rune: dict[str, int], # temp. need to see if a Rune object is necessary
                 spells: list,
                 stats: Stats,
                 tier_info: Tier) -> None:
        self._summoner = summoner
        self._participant_id = participant_id
        self._champion_id = champion_id
        self._team_key = team_key
        self._position = position
        self._role = role
        self._items = items
        self._trinket_item = trinket_item
        self._rune = rune
        self._spells = spells
        self._stats = stats
        self._tier_info = tier_info
    
    @property
    def summoner(self) -> 'Summoner':
        """
        A `Summoner` object representing a summoner
        """
        return self._summoner
    
    @summoner.setter
    def summoner(self, value: 'Summoner') -> None:
        self._summoner = value
    
    @property
    def participant_id(self) -> int:
        """
        An `int` representing the participant id
        """
        return self._participant_id
    
    @participant_id.setter
    def participant_id(self, value: int) -> None:
        self._participant_id = value
        
    @property
    def champion_id(self) -> int:
        """
        An `int` representing the champion id
        """
        return self._champion_id
    
    @champion_id.setter
    def champion_id(self, value: int) -> None:
        self._champion_id = value
    
    @property
    def team_key(self) -> str:
        """
        A `str` representing the team (BLUE, RED)
        """
        return self._team_key
    
    @team_key.setter
    def team_key(self, value: str) -> None:
        self._team_key = value
    
    @property
    def position(self) -> str:
        """
        A `str` representing the position the participant was playing
        """
        return self._position
    
    @position.setter
    def position(self, value: str) -> None:
        self._position = value
    
    @property
    def role(self) -> str:
        """
        A `str` representing the role the participant was playing
        """
        return self._role
    
    @role.setter
    def role(self, value: str) -> None:
        self._role = value
    
    @property
    def items(self) -> list[int]:
        """
        A `list[int]` representing the participants items
        """
        return self._items
    
    @items.setter
    def items(self, value: list[int]) -> None:
        self._items = value
    
    @property
    def trinket_item(self) -> int:
        """
        An `int` representing the trinket item
        """
        return self._trinket_item
    
    @trinket_item.setter
    def trinket_item(self, value: int) -> None:
        self._trinket_item = value
    
    @property
    def rune(self) -> dict[str, int]:
        """
        A `dict[str, int]` representing the rune configuration of the participant
        """
        return self._rune
    
    @rune.setter
    def rune(self, value: dict[str, int]) -> None:
        self._rune = value
    
    @property
    def spells(self) -> list:
        """
        A `list` representing the spell configuration of the participant
        """
        return self._spells
    
    @spells.setter
    def spells(self, value: list) -> None:
        self._spells = value
    
    @property
    def stats(self) -> Stats:
        """
        A `Stats` object representing the participants game stats
        """
        return self._stats
    
    @stats.setter
    def stats(self, value: Stats) -> None:
        self._stats = value
    
    @property
    def tier_info(self) -> Tier:
        """
        A `Tier` object representing the participants league tier
        """
        return self._tier_info
    
    @tier_info.setter
    def tier_info(self, value: Tier) -> None:
        self._tier_info = value
    

class Game:
    """
    Represents a game played by a summoner.\n
    
    ### Properties:
        `id: str` - Internal game id\n
        `created_at: datetime` - Date & Time of game\n
        `game_map: GameMap` - Game map (Summoners Rift, Howling Abyss, etc.)\n
        `queue_info: QueueInfo` - Queue info\n
        `version: str` - Game version\n
        `game_length_second: int` - Length of game in seconds\n
        `is_remake: bool` - Whether or not the game was a remake\n
        `is_opscore_active: bool` - Whether or not the opscore was active\n
        `is_recorded: bool` - Whether or not the game was recorded\n
        `record_info: Any` - Unknown return value and type\n
        `average_tier_info: Tier` - Average Tier of the game\n
        `participants: list[Participant]` - A list of all participant in the game\n
        `teams: list[Team]` - A list of teams in the game\n
        `memo: Any` - Unknown return value and type\n
        `myData: Participant` - User specific participant data
    """
    def __init__(self,
                 id: str,
                 created_at: datetime,
                 game_map: str,
                 queue_info: QueueInfo,
                 version: str,
                 game_length_second: int,
                 is_remake: bool,
                 is_opscore_active: bool,
                 is_recorded: bool,
                 record_info: Any,
                 average_tier_info: Tier,
                 participants: list[Participant],
                 teams: list[Team],
                 memo: Any,
                 myData: Participant) -> None:
        self._id = id
        self._created_at = created_at
        self._game_map = game_map
        self._queue_info = queue_info
        self._version = version
        self._game_length_second = game_length_second
        self._is_remake = is_remake
        self._is_opscore_active = is_opscore_active
        self._is_recorded = is_recorded
        self._record_info = record_info
        self._average_tier_info = average_tier_info
        self._participants = participants
        self._teams = teams
        self._memo = memo
        self._myData = myData
        
    @property
    def id(self) -> str:
        """
        A `str` representing the Internal game id
        """
        return self._id
    
    @id.setter
    def id(self, value: str) -> None:
        self._id = value
        
    @property
    def created_at(self) -> datetime:
        """
        A `datetime` object representing the Date & Time of game played
        """
        return self._created_at
    
    @created_at.setter
    def created_at(self, value: datetime) -> None:
        self._created_at = value
        
    @property
    def game_map(self) -> str:
        """
        A `str` representing the game map (Summoners Rift, Howling Abyss, etc.)
        """
        return self._game_map
    
    @game_map.setter
    def game_map(self, value: str) -> None:
        self._game_map = value
        
    @property
    def queue_info(self) -> QueueInfo:
        """
        A `QueueInfo` object representing queue information
        """
        return self._queue_info
    
    @queue_info.setter
    def queue_info(self, value: QueueInfo) -> None:
        self._queue_info = value
        
    @property
    def version(self) -> str:
        """
        A `str` representing the game version
        """
        return self._version
    
    @version.setter
    def version(self, value: str) -> None:
        self._version = value
        
    @property
    def game_length_second(self) -> int:
        """
        An `int` representing the length of the game in seconds
        """
        return self._game_length_second
     
    @game_length_second.setter
    def game_length_second(self, value: int) -> None:
        self._game_length_second = value
        
    @property
    def is_remake(self) -> bool:
        """
        A `bool` representing whether the game was a remake
        """
        return self._is_remake
    
    @is_remake.setter
    def is_remake(self, value: bool) -> None:
        self._is_remake = value
        
    @property
    def is_opscore_active(self) -> bool:
        """
        A `bool` representing whether the opscore was active
        """
        return self._is_opscore_active
    
    @is_opscore_active.setter
    def is_opscore_active(self, value: bool) -> None:
        self._is_opscore_active = value
        
        
    @property
    def is_recorded(self) -> bool:
        """
        A `bool` representing whether the game was recorded
        """
        return self._is_recorded
    
    @is_recorded.setter
    def is_recorded(self, value: bool) -> None:
        self._is_recorded = value
        
    @property
    def record_info(self) -> Any:
        """
        An `Any` object with unknown return value and type
        """
        return self._record_info
    
    @record_info.setter
    def record_info(self, value: Any) -> None:
        self._record_info = value
        
    @property
    def average_tier_info(self) -> Tier:
        """
        A `Tier` object representing the average league tier
        """
        return self._average_tier_info
    
    @average_tier_info.setter
    def average_tier_info(self, value: Tier) -> None:
        self._average_tier_info = value
         
    @property
    def participants(self) -> list[Participant]:
        """
        A `list[Participant]` object(s) representing all participants of a game
        """
        return self._participants
    
    @participants.setter
    def participants(self, value: list[Participant]) -> None:
        self._participants = value
    
    @property
    def teams(self) -> list[Team]:
        """
        A `list[Team]` object(s) representing all teams of a game
        """
        return self._teams
    
    @teams.setter
    def teams(self, value: list[Team]) -> None:
        self._teams = value
        
    @property
    def memo(self) -> Any:
        """
        An `Any` object with unknown return value and type
        """
        return self._memo
    
    @memo.setter
    def memo(self, value: Any) -> None:
        self._memo = value
        
    @property
    def myData(self) -> Participant:
        """
        A `Participant` object representing personal participant record
        """
        return self._myData
    
    @myData.setter
    def myData(self, value: Participant) -> None:
        self._myData = value
    
    def __repr__(self) -> str:
        return f"Game(champion_id={self.myData.champion_id}, kill={self.myData.stats.kill}, death={self.myData.stats.death}, assist={self.myData.stats.assist}, position={self.myData.position}, result={self.myData.stats.result})"


class Summoner:
    """
    Represents a summoner.\n
    
    ### Properties:
        `id: int` - Summoner id\n
        `summoner_id: str` - Summoner id\n
        `acct_id: str` - Account id\n
        `puuid: str` - Puuid\n
        `game_name: str` - In-game username\n
        `tagline: str` - Regional identifier\n
        `name: str` - Summoner name\n
        `internal_name: str` - Internal name\n
        `profile_image_url: str` - Profile image URL\n
        `level: int` - Summoner level\n
        `updated_at: datetime` - When the summoner was last updated\n
        `renewable_at: datetime` - When the summoner can be renewed\n
        `previous_seasons: Season | list[Season]` - Previous seasons\n
        `league_stats: LeagueStats | list[LeagueStats]` - League stats\n
        `most_champions: list[ChampionStats]` - Most champions\n
        `recent_game_stats: Game | list[Game]` - Recent game stats\n
    """
    def __init__(self,
                 id: int,
                 summoner_id: str,
                 acct_id: str,
                 puuid: str,
                 game_name: str,
                 tagline: str,
                 name: str,
                 internal_name: str,
                 profile_image_url: str,
                 level: int,
                 updated_at: datetime,
                 renewable_at: datetime,
                 previous_seasons: Season | list[Season] = None,
                 league_stats: LeagueStats | list[LeagueStats] = None,
                 most_champions: list[ChampionStats] = None, 
                 recent_game_stats: Game | list[Game] = None) -> None:
        self._id = id
        self._summoner_id = summoner_id
        self._acct_id = acct_id
        self._puuid = puuid
        self._game_name = game_name
        self._tagline = tagline
        self._name = name
        self._internal_name = internal_name
        self._profile_image_url = profile_image_url
        self._level = level
        self._updated_at = updated_at
        self._renewable_at = renewable_at
        self._previous_seasons = previous_seasons
        self._league_stats = league_stats
        self._most_champions = most_champions  
        self._recent_game_stats = recent_game_stats
        
    @property
    def id(self) -> int:
        """
        An `int` representing the id (opgg specific)
        """
        return self._id
    
    @id.setter
    def id(self, value: int) -> None:
        self._id = value
        
    @property
    def summoner_id(self) -> str:
        """
        A `str` representing the summoner id (riot api)
        """
        return self._summoner_id
    
    @summoner_id.setter
    def summoner_id(self, value: str) -> None:
        self._summoner_id = value
        
    @property
    def acct_id(self) -> str:
        """
        A `str` representing the account id (riot api)
        """
        return self._acct_id
    
    @acct_id.setter
    def acct_id(self, value: str) -> None:
        self._acct_id = value
        
    @property
    def puuid(self) -> str:
        """
        A `str` representing the puuid (riot api)
        """
        return self._puuid
    
    @puuid.setter
    def puuid(self, value: str) -> None:
        self._puuid = value
    
    @property
    def game_name(self) -> str:
        """
        A `str` representing the in-game username
        """
        return self._game_name
    
    @game_name.setter
    def game_name(self, value: str) -> None:
        self._game_name = value
    
    @property
    def tagline(self) -> str:
        """
        A `str` representing the tagline
        """
        return self._tagline
    
    @tagline.setter
    def tagline(self, value: str) -> None:
        self._tagline = value
    
    @property
    def name(self) -> str:
        """
        A `str` representing the summoner name
        """
        return self._name
    
    @name.setter
    def name(self, value: str) -> None:
        self._name = value
        
    @property
    def internal_name(self) -> str:
        """
        A `str` representing the internal name
        """
        return self._internal_name
    
    @internal_name.setter
    def internal_name(self, value: str) -> None:
        self._internal_name = value
        
    @property
    def profile_image_url(self) -> str:
        """
        A `str` representing the profile image url
        """
        return self._profile_image_url
    
    @profile_image_url.setter
    def profile_image_url(self, value: str) -> None:
        self._profile_image_url = value
        
    @property
    def level(self) -> int:
        """
        An `int` representing the summoner level
        """
        return self._level
    
    @level.setter
    def level(self, value: int) -> None:
        self._level = value
        
    @property
    def updated_at(self) -> datetime:
        """
        A `datetime` object representing when the summoner was last updated
        """
        return self._updated_at
    
    @updated_at.setter
    def updated_at(self, value: datetime) -> None:
        self._updated_at = value
        
    @property
    def renewable_at(self) -> datetime:
        """
        A `datetime` object representing when the summoner can be renewed
        """
        return self._renewable_at
    
    @renewable_at.setter
    def renewable_at(self, value: datetime) -> None:
        self._renewable_at = value
        
    @property
    def previous_seasons(self) -> Season | list[Season]:
        """
        A `Season | list[Season]` object(s) representing the previously played seasons
        """
        return self._previous_seasons
      
    @previous_seasons.setter
    def previous_seasons(self, value: Season | list[Season]) -> None:
        self._previous_seasons = value
        
    @property
    def league_stats(self) -> LeagueStats | list[LeagueStats]:
        """
        A `LeagueStats | list[LeagueStats]` object(s) representing all applicable league stats
        """
        return self._league_stats
    
    @league_stats.setter
    def league_stats(self, value: LeagueStats | list[LeagueStats]) -> None:
        self._league_stats = value
        
    @property
    def most_champions(self) -> list[ChampionStats]:
        """
        A `list[ChampionStats]` objects representing the most played champions
        """
        return self._most_champions
    
    @most_champions.setter
    def most_champions(self, value: list[ChampionStats]) -> None:
        self._most_champions = value
        
    @property
    def recent_game_stats(self) -> Game | list[Game]:
        """
        A `Game | list[Game]` object(s) representing the most recent games
        """
        return self._recent_game_stats
    
    @recent_game_stats.setter
    def recent_game_stats(self, value: Game | list[Game]) -> None:
        self._recent_game_stats = value

    def get_tier_from_queue(self, queue: Queue = Queue.SOLO) -> Tier:
        """
        A method to get the summoners current tier in a given queue type.
        
        ### Args:
            queue : `Queue`
                The queue type to get the tier from. Defaults to `Queue.SOLO`.
        
        ### Returns:
            `Tier` : The tier object which contains the rank, division, and lp.
        """
        
        league_stat: LeagueStats
        for league_stat in self.league_stats:
            if league_stat.queue_info.game_type == queue:
                return league_stat.tier_info
        else:
            return None
    
    def get_top_champ(self) -> ChampionStats:
        return self.most_champions[0]
        
    
    def __repr__(self) -> str:
        previous_seasons_fmt, league_stats_fmt, champion_stats_fmt, game_fmt = "", "", "", ""
        
        for season in self.previous_seasons: previous_seasons_fmt += f"{''.ljust(LJF+RJF)}  | {season}\n"
        for league_stat in self.league_stats: league_stats_fmt += f"{''.ljust(LJF+RJF)}  | {league_stat}\n"
        for champ_stat in self.most_champions: champion_stats_fmt += f"{''.ljust(LJF+RJF)}  | {champ_stat}\n"
        for game in self.recent_game_stats: game_fmt += f"{''.ljust(LJF+RJF)}  | {game}\n"
        
        return  f"[Summoner: {self.game_name}]\n{'-' * 80}\n" \
                f"{'Id'.ljust(LJF)} {'(int)'.rjust(RJF)} | {self.id}\n" \
                f"{'Summoner Id'.ljust(LJF)} {'(str)'.rjust(RJF)} | {self.summoner_id}\n" \
                f"{'Account Id'.ljust(LJF)} {'(str)'.rjust(RJF)} | {self.acct_id}\n" \
                f"{'Puuid'.ljust(LJF)} {'(str)'.rjust(RJF)} | {self.puuid}\n" \
                f"{'Game Name'.ljust(LJF)} {'(str)'.rjust(RJF)} | {self.game_name}\n" \
                f"{'Tagline'.ljust(LJF)} {'(str)'.rjust(RJF)} | {self.tagline}\n" \
                f"{'Name'.ljust(LJF)} {'(str)'.rjust(RJF)} | {self.name}\n" \
                f"{'Internal Name'.ljust(LJF)} {'(str)'.rjust(RJF)} | {self.internal_name}\n" \
                f"{'Profile Image Url'.ljust(LJF)} {'(str)'.rjust(RJF)} | {self.profile_image_url}\n" \
                f"{'Level'.ljust(LJF)} {'(int)'.rjust(RJF)} | {self.level}\n" \
                f"{'Updated At'.ljust(LJF)} {'(datetime)'.rjust(RJF)} | {self.updated_at}\n" \
                f"{'Renewable At'.ljust(LJF)} {'(datetime)'.rjust(RJF)} | {self.renewable_at}\n" \
                f"{'Previous Seasons'.ljust(LJF)} {'(Season)'.rjust(RJF)} | [List ({len(self.previous_seasons)})] \n{previous_seasons_fmt}" \
                f"{'League Stats'.ljust(LJF)} {'(LeagueStats)'.rjust(RJF)} | [List ({len(self.league_stats)})] \n{league_stats_fmt}" \
                f"{'Most Champions'.ljust(LJF)} {'(ChampStats)'.rjust(RJF)} | [List ({len(self.most_champions)})] \n{champion_stats_fmt}" \
                f"{'Recent Game Stats'.ljust(LJF)} {'(Game)'.rjust(RJF)} | [List ({len(self.recent_game_stats)})] \n{game_fmt}"